#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                $Date$
#                $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import abc

ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})
# pylint: disable=too-many-public-methods
class IGXNative(ABC):
    """
    This class is used to communicate with native serial port class. This is
    reserved for inner use. DO NOT USE.
    """

    @abc.abstractmethod
    def getPortNames(self):
        """Returns available serial ports."""

    @abc.abstractmethod
    def open(self, port):
        """
        Open serial port.

        port: Name of serial port.
        """

    @abc.abstractmethod
    def close(self):
        """
        Close serial port.
        """

    @abc.abstractmethod
    def getBaudRate(self):
        """
        Get baud rate.
        """

    @abc.abstractmethod
    def setBaudRate(self, value):
        """
        Set baud rate.
        """

    @abc.abstractmethod
    def getDataBits(self):
        """Get data bits.
        """

    @abc.abstractmethod
    def setDataBits(self, value):
        """Set amount of data bits.
        value : Amount of data bits.
        """

    @abc.abstractmethod
    def getParity(self):
        """Get parity.
        """

    @abc.abstractmethod
    def setParity(self, value):
        """
        Set parity.
        value : parity.
        """

    @abc.abstractmethod
    def getStopBits(self):
        """
        Get stop bits.
        """

    @abc.abstractmethod
    def setStopBits(self, value):
        """
        Set stop bits.
        value: Amount of stop bits.
        """

    @abc.abstractmethod
    def setBreakState(self, value):
        """
        Set break state.
        value : Is serial port in break state.
        """

    @abc.abstractmethod
    def getRtsEnable(self):
        """
        Get Request To Send state.
        """

    @abc.abstractmethod
    def setRtsEnable(self, value):
        """Set Request To Send state.
        value: Is RTS set.
        """

    @abc.abstractmethod
    def getDtrEnable(self):
        """
        Is Data Terminal ready set.
        """

    @abc.abstractmethod
    def setDtrEnable(self, value):
        """Is Data Terminal ready set.
        value : True, if DTR is set.
        """

    @abc.abstractmethod
    def getDsrHolding(self):
        """
        Get Get Data Set Ready holding flag.
        """

    @abc.abstractmethod
    def getBytesToRead(self):
        """
        Returns amount of bytes to read.
        """

    @abc.abstractmethod
    def getBytesToWrite(self):
        """
        Returns amount of bytes to write.
        """

    @abc.abstractmethod
    def read(self):
        """Read data from serial port to the buffer."""

    @abc.abstractmethod
    def write(self, data):
        """Write data to the serial port."""

    @abc.abstractmethod
    def getCtsHolding(self):
        """Returns Clear To Send holding flag."""

    @abc.abstractmethod
    def getCDHolding(self):
        """Gets the state of the Carrier Detect line for the port."""


    @abc.abstractmethod
    def getHandshake(self):
        """Gets the handshaking protocol for serial port transmission of data."""

    @abc.abstractmethod
    def setHandshake(self, value):
        """Sets the handshaking protocol for serial port transmission of data."""

    @abc.abstractmethod
    def isOpen(self):
        """Is connnection open."""



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                $Date$
#                $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import ctypes
import ctypes.wintypes
import threading
from .GXSettings import GXSettings
from .IGXNative import IGXNative

#Constant values.
INVALID_HANDLE_VALUE = -1
MS_DSR_ON = 32
EV_RING = 256
EV_PERR = 512
EV_ERR = 128
SETXOFF = 1
EV_RXCHAR = 1
OPEN_EXISTING = 3
FILE_ATTRIBUTE_NORMAL = 128
FILE_FLAG_OVERLAPPED = 1073741824
GENERIC_WRITE = 1073741824
GENERIC_READ = 2147483648
PURGE_TXABORT = 1
PURGE_RXABORT = 2
PURGE_TXCLEAR = 4
PURGE_RXCLEAR = 8

RTS_CONTROL_HANDSHAKE = 2
RTS_CONTROL_DISABLE = 0
RTS_CONTROL_ENABLE = 1
RTS_CONTROL_TOGGLE = 3

DTR_CONTROL_HANDSHAKE = 2
DTR_CONTROL_DISABLE = 0
DTR_CONTROL_ENABLE = 1
SETRTS = 3
CLRRTS = 4
SETDTR = 5
CLRDTR = 6
XON = 0x11
XOFF = 0x13
MS_RLSD_ON = 128
#Error code defs.
ERROR_SUCCESS = 0
ERROR_NOT_ENOUGH_MEMORY = 8
ERROR_OPERATION_ABORTED = 995
ERROR_IO_INCOMPLETE = 996
ERROR_IO_PENDING = 997
ERROR_INVALID_USER_BUFFER = 1784

#Error code defs.
ERROR_SUCCESS = 0
ERROR_NOT_ENOUGH_MEMORY = 8
ERROR_OPERATION_ABORTED = 995
ERROR_IO_INCOMPLETE = 996
ERROR_IO_PENDING = 997
ERROR_INVALID_USER_BUFFER = 1784

READ_CONTROL = 0x00020000
STANDARD_RIGHTS_READ = READ_CONTROL

KEY_QUERY_VALUE = 0x0001
HKEY_LOCAL_MACHINE = 0x80000002
SYNCHRONIZE = 0x00100000
KEY_ENUMERATE_SUB_KEYS = 0x0008
KEY_NOTIFY = 0x0010
KEY_READ = ((STANDARD_RIGHTS_READ | KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS | KEY_NOTIFY) & (~SYNCHRONIZE))
KEY_EXECUTE = KEY_READ & ~SYNCHRONIZE

if ctypes.sizeof(ctypes.c_void_p) == 8:
    ULONG_PTR = ctypes.c_int64
else:
    ULONG_PTR = ctypes.c_ulong

#Structyres.
class DUMMYSTRUCTNAME(ctypes.Structure):
    # pylint: disable=too-few-public-methods
    _fields_ = [('Offset', ctypes.wintypes.DWORD),\
    ('OffsetHigh', ctypes.wintypes.DWORD),]

class DUMMYUNIONNAME(ctypes.Union):
    # pylint: disable=too-few-public-methods
    _fields_ = [('_0', DUMMYSTRUCTNAME),\
    ('Pointer', ctypes.c_void_p),]

class OVERLAPPED(ctypes.Structure):
    # pylint: disable=too-few-public-methods
    _fields_ = [('Internal', ULONG_PTR),\
    ('InternalHigh', ULONG_PTR),\
    ('_0', DUMMYUNIONNAME),\
    ('hEvent', ctypes.wintypes.HANDLE),]

class COMSTAT(ctypes.Structure):
    # pylint: disable=too-few-public-methods
    _fields_ = [('fCtsHold', ctypes.wintypes.DWORD, 1),\
    ('fDsrHold', ctypes.wintypes.DWORD, 1),\
    ('fRlsdHold', ctypes.wintypes.DWORD, 1),\
    ('fXoffHold', ctypes.wintypes.DWORD, 1),\
    ('fXoffSent', ctypes.wintypes.DWORD, 1),\
    ('fEof', ctypes.wintypes.DWORD, 1),\
    ('fTxim', ctypes.wintypes.DWORD, 1),\
    ('fReserved', ctypes.wintypes.DWORD, 25),\
    ('cbInQue', ctypes.wintypes.DWORD),\
    ('cbOutQue', ctypes.wintypes.DWORD),]

class DCB(ctypes.Structure):
    # pylint: disable=too-few-public-methods, too-many-instance-attributes
    _fields_ = [('DCBlength', ctypes.wintypes.DWORD),\
    ('BaudRate', ctypes.wintypes.DWORD),\
    ('fBinary', ctypes.wintypes.DWORD, 1),\
    ('fParity', ctypes.wintypes.DWORD, 1),\
    ('fOutxCtsFlow', ctypes.wintypes.DWORD, 1),\
    ('fOutxDsrFlow', ctypes.wintypes.DWORD, 1),\
    ('fDtrControl', ctypes.wintypes.DWORD, 2),\
    ('fDsrSensitivity', ctypes.wintypes.DWORD, 1),\
    ('fTXContinueOnXoff', ctypes.wintypes.DWORD, 1),\
    ('fOutX', ctypes.wintypes.DWORD, 1),\
    ('fInX', ctypes.wintypes.DWORD, 1),\
    ('fErrorChar', ctypes.wintypes.DWORD, 1),\
    ('fNull', ctypes.wintypes.DWORD, 1),\
    ('fRtsControl', ctypes.wintypes.DWORD, 2),\
    ('fAbortOnError', ctypes.wintypes.DWORD, 1),\
    ('fDummy2', ctypes.wintypes.DWORD, 17),\
    ('wReserved', ctypes.wintypes.WORD),\
    ('XonLim', ctypes.wintypes.WORD),\
    ('XoffLim', ctypes.wintypes.WORD),\
    ('ByteSize', ctypes.wintypes.BYTE),\
    ('Parity', ctypes.wintypes.BYTE),\
    ('StopBits', ctypes.wintypes.BYTE),\
    ('XonChar', ctypes.c_char),\
    ('XoffChar', ctypes.c_char),\
    ('ErrorChar', ctypes.c_char),\
    ('EofChar', ctypes.c_char),\
    ('EvtChar', ctypes.c_char),\
    ('wReserved1', ctypes.wintypes.WORD),]

#ctypes.windll.kernel32.WriteFile don't work for all Windows versions for some reason.
_stdcall_libraries = {}
_stdcall_libraries['kernel32'] = ctypes.WinDLL('kernel32')
WriteFile = _stdcall_libraries['kernel32'].WriteFile
WriteFile.restype = ctypes.wintypes.BOOL
LPOVERLAPPED = ctypes.POINTER(OVERLAPPED)
# pylint: disable=no-member

#LPDWORD is missing from older ctypes.wintypes.
LPDWORD = ctypes.POINTER(ctypes.wintypes.DWORD)

WriteFile.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.LPCVOID, ctypes.wintypes.DWORD, LPDWORD, LPOVERLAPPED]

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class GXWindowsHandler(GXSettings, IGXNative):
    def __init__(self):
        """Constructor."""
        GXSettings.__init__(self)
        self.h = INVALID_HANDLE_VALUE
        self._overlapped_read = None
        self._overlapped_write = None
        self._closing = ctypes.windll.Kernel32.CreateEventW(0, 0, 0, 0)
        self.__closed = threading.Event()

    def __del__(self):
        """Destructor."""
        if self._closing != INVALID_HANDLE_VALUE:
            ctypes.windll.Kernel32.CloseHandle(self._closing)
            self._closing = INVALID_HANDLE_VALUE

    def getPortNames(self):
        """Returns available serial ports."""
        #Use RegOpenKeyEx() with the new Registry path to get an open handle
        #to the child key you want to enumerate.
        hKey = ctypes.wintypes.HKEY()
        ret = ctypes.windll.Kernel32.RegOpenKeyExW(HKEY_LOCAL_MACHINE, "HARDWARE\\DEVICEMAP\\SERIALCOMM", \
            0, KEY_ENUMERATE_SUB_KEYS | KEY_EXECUTE | KEY_QUERY_VALUE, ctypes.byref(hKey))
        if ret != 0:
            raise Exception('Failed to get port names: {!r}'.format(ctypes.WinError(ret)))

        dwType = ctypes.wintypes.DWORD()
        dwcValueName = ctypes.wintypes.DWORD(256)
        valueName = ctypes.create_unicode_buffer(256)
        cbData = ctypes.wintypes.DWORD(256)
        deviceName = ctypes.create_unicode_buffer(236)
        pos = 0
        ports = []
        while pos < 100:
            ret = ctypes.windll.Kernel32.RegEnumValueW(hKey, pos, valueName, ctypes.byref(dwcValueName), None, ctypes.byref(dwType), deviceName, ctypes.byref(cbData))
            if ret != 0:
                break
            ports.append(deviceName.value)
            pos += 1
        ctypes.windll.Kernel32.RegCloseKey(hKey)
        return ports

    def __getCommState(self):
        if self.h == INVALID_HANDLE_VALUE:
            raise Exception("Serial port is not open")
        ctypes.windll.Kernel32.SetCommMask(self.h, EV_ERR)
        # Get state
        dcb = DCB()
        if ctypes.windll.Kernel32.GetCommState(self.h, ctypes.byref(dcb)) == 0:
            raise Exception("Failed to get comm state.")
        return dcb

    def __updateSettings(self):
        """Set communication parameters."""
        # pylint: disable=attribute-defined-outside-init
        if self.h == INVALID_HANDLE_VALUE:
            raise Exception("Serial port is not open")

        dcb = self.__getCommState()
        dcb.BaudRate = self.baudrate
        dcb.ByteSize = self.dataBits
        dcb.Parity = self.parity
        # Disable Parity Check
        if dcb.Parity == 0:
            dcb.fParity = 0
        else:
            dcb.fParity = 1
        dcb.StopBits = self.stopBits
        dcb.fBinary = 1
        dcb.fNull = 0
        dcb.fErrorChar = 0
        dcb.fAbortOnError = 0
        dcb.XonChar = XON
        dcb.XoffChar = XOFF
        dcb.fRtsControl = 1
        dcb.fDtrControl = 1
        if not ctypes.windll.Kernel32.SetCommState(self.h, ctypes.byref(dcb)):
            raise Exception('Failed to set serial port settings: {!r}'.format(ctypes.WinError()))

    def open(self, port):
        #pylint: disable=bare-except
        """
        Open serial port.

        port: Name of serial port.
        """
        self.close()
        if not port:
            raise Exception("Invalid serial port name.")
        # Open the file for writing.
        self.h = ctypes.windll.Kernel32.CreateFileW('\\\\.\\' + port, GENERIC_READ | GENERIC_WRITE,
                                                    0,  # exclusive access
                                                    None,  # no security
                                                    OPEN_EXISTING,
                                                    FILE_FLAG_OVERLAPPED,
                                                    0)
        if self.h == INVALID_HANDLE_VALUE:
            ret = ctypes.windll.Kernel32.GetLastError()
            raise Exception("Failed to open port {!r}: {!r}".format(port, ctypes.WinError(ret)))
        try:
            self._overlapped_read = OVERLAPPED(hEvent=ctypes.windll.Kernel32.CreateEventW(0, 0, 0, 0))
            self._overlapped_write = OVERLAPPED(hEvent=ctypes.windll.Kernel32.CreateEventW(0, 0, 0, 0))
            if ctypes.windll.Kernel32.ResetEvent(self._closing) == 0:
                ret = ctypes.windll.Kernel32.GetLastError()
                raise Exception('Failed to open port: {!r}'.format(ctypes.WinError(ret)))
            self.__updateSettings()
            #Clear buffers.
            ctypes.windll.Kernel32.PurgeComm(self.h, PURGE_TXCLEAR | PURGE_TXABORT | PURGE_RXCLEAR | PURGE_RXABORT)
        except Exception as e:
            try:
                self.close()
            except:
                pass
            raise e

    def isOpen(self):
        return self.h != INVALID_HANDLE_VALUE

    def close(self):
        """
        Close serial port.
        """
        if self.h != INVALID_HANDLE_VALUE:
            ctypes.windll.Kernel32.SetEvent(self._closing)
            self.__closed.wait()
            if self._overlapped_read:
                ctypes.windll.Kernel32.CloseHandle(self._overlapped_read.hEvent)
                self._overlapped_read = None
            if self._overlapped_write:
                ctypes.windll.Kernel32.CloseHandle(self._overlapped_write.hEvent)
                self._overlapped_write = None
            ctypes.windll.Kernel32.CloseHandle(self.h)
            self.h = INVALID_HANDLE_VALUE

    def getBaudRate(self):
        """
        Get baud rate.
        """
        return self.__getCommState().BaudRate


    def setBaudRate(self, value):
        """
        Set baud rate.
        """
        self.__updateSettings()


    def getDataBits(self):
        """Get data bits.
        """
        return self.__getCommState().ByteSize


    def setDataBits(self, value):
        """Set amount of data bits.
        value : Amount of data bits.
        """
        self.__updateSettings()


    def getParity(self):
        """Get parity.
        """
        return self.__getCommState().fParity

    def setParity(self, value):
        """
        Set parity.
        value : parity.
        """
        self.__updateSettings()


    def getStopBits(self):
        """
        Get stop bits.
        """
        return self.__getCommState().StopBits


    def setStopBits(self, value):
        """
        Set stop bits.
        value: Amount of stop bits.
        """
        self.__updateSettings()

    def setBreakState(self, value):
        """
        Set break state.
        value : Is serial port in break state.
        """
        if value:
            if ctypes.windll.Kernel32.SetCommBreak(self.h) == 0:
                raise Exception("setBreakState failed ({!r})".format(ctypes.WinError()))
        else:
            if ctypes.windll.Kernel32.ClearCommBreak(self.h) == 0:
                raise Exception("setBreakState failed ({!r})".format(ctypes.WinError()))

    def getRtsEnable(self):
        """
        Get Request To Send state.
        """
        dcb = self.__getCommState()
        return dcb.fRtsControl == RTS_CONTROL_ENABLE

    def setRtsEnable(self, value):
        """Set Request To Send state.
        value: Is RTS set.
        """
        if value:
            tmp = ctypes.wintypes.DWORD(SETRTS)
        else:
            tmp = ctypes.wintypes.DWORD(CLRRTS)
        if ctypes.windll.Kernel32.EscapeCommFunction(self.h, tmp) == 0:
            raise Exception("setRtsEnable failed ({!r})".format(ctypes.WinError()))

    def getDtrEnable(self):
        """
        Is Data Terminal ready set.
        """
        dcb = self.__getCommState()
        return dcb.fDtrControl == DTR_CONTROL_ENABLE

    def setDtrEnable(self, value):
        """Is Data Terminal ready set.
        value : True, if DTR is set.
        """
        if value:
            tmp = ctypes.wintypes.DWORD(SETDTR)
        else:
            tmp = ctypes.wintypes.DWORD(CLRDTR)
        if ctypes.windll.Kernel32.EscapeCommFunction(self.h, tmp) == 0:
            raise Exception("Read failed ({!r})".format(ctypes.WinError()))

    def getDsrHolding(self):
        """
        Get Get Data Set Ready holding flag.
        """
        flags = ctypes.wintypes.DWORD()
        comstat = COMSTAT()
        if not ctypes.windll.Kernel32.ClearCommError(self.h, ctypes.byref(flags), ctypes.byref(comstat)):
            raise Exception("ClearCommError failed ({!r})".format(ctypes.WinError()))
        return comstat.fDsrHold != 0

    def getBytesToRead(self):
        """
        Returns amount of bytes to read.
        """
        flags = ctypes.wintypes.DWORD()
        comstat = COMSTAT()
        if not ctypes.windll.Kernel32.ClearCommError(self.h, ctypes.byref(flags), ctypes.byref(comstat)):
            raise Exception("ClearCommError failed ({!r})".format(ctypes.WinError()))
        return comstat.cbInQue

    def getBytesToWrite(self):
        """
        Returns amount of bytes to write.
        """
        flags = ctypes.wintypes.DWORD()
        comstat = COMSTAT()
        if not ctypes.windll.ClearCommError(self.h, ctypes.byref(flags), ctypes.byref(comstat)):
            raise Exception("ClearCommError failed ({!r})".format(ctypes.WinError()))
        return comstat.cbOutQue

    def read(self):
        """Read data from serial port to the buffer."""
        if self.h == INVALID_HANDLE_VALUE:
            raise Exception("Serial port is not open")
        count = self.getBytesToRead()
        if count == 0:
            count = 1
        buf = ctypes.create_string_buffer(count)
        c = ctypes.wintypes.DWORD()
        if ctypes.windll.Kernel32.ReadFile(self.h, buf, count, ctypes.byref(c), ctypes.byref(self._overlapped_read)) == 0:
            ret = ctypes.windll.Kernel32.GetLastError()
            if ret not in (ERROR_SUCCESS, ERROR_IO_PENDING):
                raise Exception("Read failed ({!r})".format(ctypes.WinError()))
            arrtype = ctypes.wintypes.HANDLE * 2
            handle_array = arrtype(self._closing, self._overlapped_read.hEvent)
            ret = ctypes.windll.kernel32.WaitForMultipleObjects(2, handle_array, 0, -1)
            #If user has close the media.
            if ret == 0:
                self.__closed.set()
                return None

            if ctypes.windll.Kernel32.GetOverlappedResult(self.h, ctypes.byref(self._overlapped_read), ctypes.byref(c), True) == 0:
                if ctypes.windll.Kernel32.GetLastError() != ERROR_OPERATION_ABORTED:
                    raise Exception("Read failed  ({!r})".format(ctypes.WinError()))
        return bytearray(buf.raw[:c.value])

    def write(self, data):
        """Write data to the serial port."""
        if self.h == INVALID_HANDLE_VALUE:
            raise Exception("Serial port is not open")
        n = ctypes.wintypes.DWORD()
        if WriteFile(self.h, data, len(data), ctypes.byref(n), self._overlapped_write) == 0:
            errorcode = ERROR_SUCCESS
        else:
            errorcode = ctypes.windll.Kernel32.GetLastError()
        if errorcode in (ERROR_INVALID_USER_BUFFER, ERROR_NOT_ENOUGH_MEMORY, ERROR_OPERATION_ABORTED):
            return 0
        if errorcode == ERROR_IO_PENDING:
            arrtype = ctypes.wintypes.HANDLE * 2
            handle_array = arrtype(self._closing, self._overlapped_write.hEvent)
            ret = ctypes.windll.kernel32.WaitForMultipleObjects(2, handle_array, 0, 1000)
            #If user has close the media.
            if ret == 0:
                return None
        if errorcode == ERROR_SUCCESS:
            return len(data)
        raise Exception("write failed ({!r})".format(ctypes.WinError()))

    def getCtsHolding(self):
        """Returns Clear To Send holding flag."""
        flags = ctypes.wintypes.DWORD()
        comstat = COMSTAT()
        if not ctypes.windll.Kernel32.ClearCommError(self.h, ctypes.byref(flags), ctypes.byref(comstat)):
            raise Exception("ClearCommError failed ({!r})".format(ctypes.WinError()))
        return comstat.fCtsHold != 0

    def getCDHolding(self):
        """Gets the state of the Carrier Detect line for the port."""
        status = ctypes.wintypes.DWORD()
        if ctypes.windll.Kernel32.GetCommModemStatus(self.h, status) == 0:
            raise Exception("ClearCommError failed ({!r})".format(ctypes.WinError()))
        return (status & MS_RLSD_ON) != 0

    def getHandshake(self):
        """Gets the handshaking protocol for serial port transmission of data."""
        dcb = self.__getCommState()
        #Disable DTR monitoring
        #Disable RTS (Ready To Send)
        if dcb.fDtrControl == DTR_CONTROL_DISABLE and dcb.fRtsControl == DTR_CONTROL_DISABLE:
            #Enable XON/XOFF for transmission
            #Enable XON/XOFF for receiving
            if dcb.fOutX and dcb.fInX:
                #XOnXOff
                return 1
            #None
            return 0
        #Enable XON/XOFF for transmission
        #Enable XON/XOFF for receiving
        if dcb.fOutX and dcb.fInX:
            #RequestToSendXOnXOff
            return 3
        #hardware flow control is used.
        return 2

    def setHandshake(self, value):
        # pylint: disable=attribute-defined-outside-init
        """Sets the handshaking protocol for serial port transmission of data."""
        dcb = self.__getCommState()
        #None
        if value == 0:
            dcb.fDtrControl = DTR_CONTROL_DISABLE
            #Disable RTS (Ready To Send)
            dcb.fRtsControl = DTR_CONTROL_DISABLE
            dcb.fOutX = dcb.fInX = 0
        #XOnXOff
        elif value == 1:
            dcb.fDtrControl = DTR_CONTROL_DISABLE
            #Disable RTS (Ready To Send)
            dcb.fRtsControl = DTR_CONTROL_DISABLE
            dcb.fOutX = dcb.fInX = 1
        #hardware flow control is used.
        elif value == 2:
            dcb.fDtrControl = DTR_CONTROL_ENABLE
            #Disable RTS (Ready To Send)
            dcb.fRtsControl = DTR_CONTROL_ENABLE
            dcb.fOutX = dcb.fInX = 0
        #RequestToSendXOnXOff
        elif value == 3:
            dcb.fDtrControl = DTR_CONTROL_ENABLE
            #Disable RTS (Ready To Send)
            dcb.fRtsControl = DTR_CONTROL_ENABLE
            dcb.fOutX = dcb.fInX = 1
        if ctypes.windll.Kernel32.SetCommState(self.h, ctypes.byref(dcb)) == 0:
            raise Exception("Failed to set comm state.")



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
# pylint: disable=import-error
import glob
import os.path
import os
import fcntl
import array
import select
import struct
import termios
from .GXSettings import GXSettings
from .IGXNative import IGXNative

#Constant values.
_CMSPAR = 0o10000000000
_TIOCSBRK = 0x5427
_TIOCCBRK = 0x5428

_BAUDRATE_CONSTANTS = {
    50: termios.B50, 75: termios.B75, 110: termios.B110, 134: termios.B134,
    150: termios.B150, 200: termios.B200, 300: termios.B300,
    600: termios.B600, 1200: termios.B1200, 1800: termios.B1800,
    2400: termios.B2400, 4800: termios.B4800, 9600: termios.B9600,
    19200: termios.B19200, 38400: termios.B38400, 57600: termios.B57600,
    115200: termios.B115200, 230400: termios.B230400,
    # Linux baudrates that are not included in termios module
    460800: 0x1004, 500000: 0x1005, 576000: 0x1006,
    921600: 0x1007, 1000000: 0x1008, 1152000: 0x1009,
    1500000: 0x100A, 2000000: 0x100B, 2500000: 0x100C,
    3000000: 0x100D, 3500000: 0x100E, 4000000: 0x100F,
    }

_DATABITS_TO_CFLAG = {5: termios.CS5, 6: termios.CS6, 7: termios.CS7, 8: termios.CS8}

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class GXLinuxHandler(GXSettings, IGXNative):
    def __init__(self):
        """Constructor."""
        GXSettings.__init__(self)
        self.h = None
        self.__closedR = None
        self.__closedW = None

    def getPortNames(self):
        """Returns available serial ports."""
        tmp = glob.glob('/dev/ttyS*')
        tmp.extend(glob.glob('/dev/ttyUSB*'))
        tmp.extend(glob.glob('/dev/ttyXRUSB*'))
        tmp.extend(glob.glob('/dev/ttyACM*'))
        tmp.extend(glob.glob('/dev/ttyAMA*'))
        tmp.extend(glob.glob('/dev/rfcomm*'))
        tmp.extend(glob.glob('/dev/ttyAP*'))
        # hide non-present internal serial ports
        devices = []
        for device in tmp:
            name = os.path.basename(device)
            if os.path.exists('/sys/class/tty/{}/device'.format(name)):
                devices.append(device)
        return devices

    def isOpen(self):
        return self.h

    def open(self, port):
        #pylint: disable=bare-except,no-member
        """
        Open serial port.

        port: Name of serial port.
        """
        self.close()
        if not port:
            raise Exception("Invalid serial port name.")
        self.h = os.open(port, os.O_RDWR | os.O_NOCTTY | os.O_NONBLOCK)
        (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) = termios.tcgetattr(self.h)
        # set up raw mode / no echo / binary
        cflag |= (termios.CLOCAL | termios.CREAD)
        lflag &= ~(termios.ICANON | termios.ECHO | termios.ECHOE | \
            termios.ECHOK | termios.ECHONL | \
            termios.ISIG | termios.IEXTEN)  # |termios.ECHOPRT
        oflag &= ~(termios.OPOST | termios.ONLCR | termios.OCRNL)
        iflag &= ~(termios.INLCR | termios.IGNCR | termios.ICRNL | termios.IGNBRK)
        #Baud rate
        ispeed = _BAUDRATE_CONSTANTS[int(self.baudrate)]
        ospeed = _BAUDRATE_CONSTANTS[int(self.baudrate)]
        #Databits
        cflag &= ~termios.CSIZE
        if self.dataBits == 8:
            cflag |= termios.CS8
        elif self.dataBits == 7:
            cflag |= termios.CS7
        elif self.dataBits == 6:
            cflag |= termios.CS6
        elif self.dataBits == 5:
            cflag |= termios.CS5
        #Stop bits
        if int(self.stopBits) == 0:
            cflag &= ~(termios.CSTOPB)
        else:
            cflag |= (termios.CSTOPB)
        # setup parity
        iflag &= ~(termios.INPCK | termios.ISTRIP)
        p = int(self.parity)
        if p == 0: #Parity.NONE
            cflag &= ~(termios.PARENB | termios.PARODD | _CMSPAR)
        elif p == 2: #Parity.PARITY_EVEN
            cflag &= ~(termios.PARODD | _CMSPAR)
            cflag |= (termios.PARENB)
        elif p == 1: #Parity.ODD:
            cflag &= ~_CMSPAR
            cflag |= (termios.PARENB | termios.PARODD)
        elif p == 3 and _CMSPAR:#Parity.MARK
            cflag |= (termios.PARENB | _CMSPAR | termios.PARODD)
        elif p == 4 and _CMSPAR:#Parity.SPACE
            cflag |= (termios.PARENB | _CMSPAR)
            cflag &= ~(termios.PARODD)

        #xonxoff off
        iflag &= ~(termios.IXON | termios.IXOFF)
        # rtscts
        cflag &= ~(termios.CRTSCTS)
        termios.tcsetattr(self.h, termios.TCSANOW, [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])
        #Clear imput buffer.
        termios.tcflush(self.h, termios.TCIFLUSH)
        self.__closedR, self.__closedW = os.pipe()
        fcntl.fcntl(self.__closedR, fcntl.F_SETFL, os.O_NONBLOCK)

    def close(self):
        """
        Close serial port.
        """
        if self.h:
            os.write(self.__closedW, b'1')
            os.close(self.__closedW)
            os.close(self.__closedR)
            self.__closedW = None
            self.__closedR = None
            os.close(self.h)
            self.h = None

    def fileno(self):
        return self.h

    def getBaudRate(self):
        """
        Get baud rate.
        """
        try:
            (_, _, _, _, _, rate, _) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("getBaudRate failed. " + str(e))
        for k, v in _BAUDRATE_CONSTANTS.items():
            if v == rate:
                return k
        raise Exception("Unknown baud rate: " + str(rate))

    def setBaudRate(self, value):
        """
        Set baud rate.
        """
        try:
            (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("setBaudRate failed. " + str(e))
        cflag &= ~(termios.CBAUD | termios.CBAUDEX)
        cflag |= _BAUDRATE_CONSTANTS[value]
        ispeed = _BAUDRATE_CONSTANTS[value]
        ospeed = _BAUDRATE_CONSTANTS[value]
        try:
            termios.tcsetattr(self.h, termios.TCSANOW, [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])
        except termios.error as e:
            raise Exception("setBaudRate failed. " + str(e))

    def getDataBits(self):
        """
        Get data bits.
        """
        try:
            (_, _, cflag, _, _, _, _) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("getDataBits failed. " + str(e))

        cs = cflag & termios.CSIZE
        for k, v in _DATABITS_TO_CFLAG.items():
            if v == cs:
                return k
        raise Exception("Unknown data bits: " + str(cs))

    def setDataBits(self, value):
        """
        Set amount of data bits.
        value : Amount of data bits.
        """
        try:
            (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("setDataBits failed. " + str(e))
        cflag &= ~termios.CSIZE
        cflag |= _DATABITS_TO_CFLAG[value]
        try:
            termios.tcsetattr(self.h, termios.TCSANOW, [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])
        except termios.error as e:
            raise Exception("setDataBits failed. " + str(e))

    def getParity(self):
        """Get parity.
        """
        try:
            (_, _, cflag, _, _, _, _) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("getParity failed. " + str(e))
        #Parity.MARK
        if (cflag & (termios.PARENB | _CMSPAR | termios.PARODD)) == (termios.PARENB | _CMSPAR | termios.PARODD):
            return 3
        #SPACE
        if (cflag & (termios.PARENB | _CMSPAR)) == (termios.PARENB | _CMSPAR):
            return 4
        #ODD
        if (cflag & (termios.PARENB | termios.PARODD)) == (termios.PARENB | termios.PARODD):
            return 1
        #EVEN
        if (cflag & (termios.PARENB)) == (termios.PARENB):
            return 2
        #NONE
        return 0

    def setParity(self, value):
        """
        Set parity.
        value : parity.
        """
        try:
            (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("setParity failed. " + str(e))

        iflag &= ~(termios.INPCK | termios.ISTRIP)
        if self.parity == 0: #Parity.NONE:
            cflag &= ~(termios.PARENB | termios.PARODD | _CMSPAR)
        elif self.parity == 2:#Parity.EVEN:
            cflag &= ~(termios.PARODD | _CMSPAR)
            cflag |= (termios.PARENB)
        elif self.parity == 1: #Parity.ODD:
            cflag &= ~_CMSPAR
            cflag |= (termios.PARENB | termios.PARODD)
        elif self.parity == 3 and _CMSPAR: #Parity.MARK
            cflag |= (termios.PARENB | _CMSPAR | termios.PARODD)
        elif self.parity == 4 and _CMSPAR: #Parity.SPACE
            cflag |= (termios.PARENB | _CMSPAR)
            cflag &= ~(termios.PARODD)
        # Set tty attributes
        try:
            termios.tcsetattr(self.h, termios.TCSANOW, [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])
        except termios.error as e:
            raise Exception("setParity failed. " + str(e))

    def getStopBits(self):
        """
        Get stop bits.
        """
        try:
            (_, _, cflag, _, _, _, _) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("getStopBits failed. " + str(e))
        if cflag & termios.CSTOPB != 0:
            return 1
        return 0

    def setStopBits(self, value):
        """
        Set stop bits.
        value: Amount of stop bits.
        """
        try:
            (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("setStopBits failed. " + str(e))
        cflag &= ~termios.CSTOPB
        if value == 1:
            cflag |= termios.CSTOPB
        try:
            termios.tcsetattr(self.h, termios.TCSANOW, [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])
        except termios.error as e:
            raise Exception("setStopBits failed. " + str(e))

    def setBreakState(self, value):
        """
        Set break state.
        value : Is serial port in break state.
        """
        if value:
            fcntl.ioctl(self.h, _TIOCSBRK)
        else:
            fcntl.ioctl(self.h, _TIOCCBRK)


    def getRtsEnable(self):
        """
        Get Request To Send state.
        """
        try:
            (_, _, cflag, _, _, _, _) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("getCtsHolding failed. " + str(e))
        return (cflag & termios.CRTSCTS) != 0

    def setRtsEnable(self, value):
        """Set Request To Send state.
        value: Is RTS set.
        """
        tmp = struct.pack('I', 4)
        if value:
            fcntl.ioctl(self.h, 0x5416, tmp)
        else:
            fcntl.ioctl(self.h, 0x5417, tmp)

    def getDtrEnable(self):
        """
        Is Data Terminal ready set.
        """
        try:
            (_, _, cflag, _, _, _, _) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("getCtsHolding failed. " + str(e))
        return cflag & termios.CRTSCTS != 0

    def setDtrEnable(self, value):
        """Is Data Terminal ready set.
        value : True, if DTR is set.
        """
        tmp = struct.pack('I', 2)
        if value:
            fcntl.ioctl(self.h, 0x5416, tmp)
        else:
            fcntl.ioctl(self.h, 0x5417, tmp)

    def getDsrHolding(self):
        """
        Get Get Data Set Ready holding flag.
        """

    def getBytesToRead(self):
        """
        Returns amount of bytes to read.
        """
        buf = array.array('I', [0])
        fcntl.ioctl(self.h, termios.TIOCINQ, buf, True)
        return buf[0]

    def getBytesToWrite(self):
        """
        Returns amount of bytes to write.
        """
        buf = array.array('I', [0])
        fcntl.ioctl(self.h, termios.TIOCOUTQ, buf, True)
        return buf[0]

    def read(self):
        """Read data from serial port to the buffer."""
        ready, _, _ = select.select([self.h, self.__closedR], [], [], )
        if self.__closedR in ready:
            return None
        cnt = self.getBytesToRead()
        if cnt == 0:
            cnt = 1
        ret = os.read(self.h, cnt)
        if ret is None:
            return None
        if isinstance(ret, int):
            return bytearray([ret])
        return bytearray(ret)

    def write(self, data):
        """Write data to the serial port."""
        ret = os.write(self.h, data)
        termios.tcflush(self.h, termios.TCOFLUSH)
        return ret

    def getCtsHolding(self):
        """Returns Clear To Send holding flag."""
        try:
            (_, _, cflag, _, _, _, _) = termios.tcgetattr(self.h)
        except termios.error as e:
            raise Exception("getCtsHolding failed. " + str(e))
        return cflag & termios.CRTSCTS != 0

    def getCDHolding(self):
        """Gets the state of the Carrier Detect line for the port."""

    def getHandshake(self):
        """Gets the handshaking protocol for serial port transmission of data."""

    def setHandshake(self, value):
        # pylint: disable=attribute-defined-outside-init
        """Sets the handshaking protocol for serial port transmission of data."""
